'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var BigInt$1 = require('./BigInt.cjs');
var Serializer = require('./Serializer.cjs');
var UnexpectedAmountError = require('./errors/UnexpectedAmountError.cjs');
var AmountMismatchError = require('./errors/AmountMismatchError.cjs');

const createAmount = (basisPoints, identifier, decimals) => ({
  basisPoints: BigInt$1.createBigInt(basisPoints),
  identifier,
  decimals
});
const createAmountFromDecimals = (decimalAmount, identifier, decimals) => {
  const exponentAmount = createAmount(BigInt(10) ** BigInt(decimals ?? 0), identifier, decimals);
  return multiplyAmount(exponentAmount, decimalAmount);
};
const tokenAmount = (tokens, identifier, decimals) => createAmountFromDecimals(tokens, identifier ?? 'Token', decimals ?? 0);
const lamports = lamports => createAmount(lamports, 'SOL', 9);
const sol = sol => createAmountFromDecimals(sol, 'SOL', 9);
const usd = usd => createAmountFromDecimals(usd, 'USD', 2);
const isAmount = (amount, identifier, decimals) => amount.identifier === identifier && amount.decimals === decimals;
const isSolAmount = amount => isAmount(amount, 'SOL', 9);
const sameAmounts = (left, right) => isAmount(left, right.identifier, right.decimals);
function assertAmount(amount, identifier, decimals) {
  if (!isAmount(amount, identifier, decimals)) {
    throw new UnexpectedAmountError.UnexpectedAmountError(amount, identifier, decimals);
  }
}
function assertSolAmount(actual) {
  assertAmount(actual, 'SOL', 9);
}
function assertSameAmounts(left, right, operation) {
  if (!sameAmounts(left, right)) {
    throw new AmountMismatchError.AmountMismatchError(left, right, operation);
  }
}
const addAmounts = (left, right) => {
  assertSameAmounts(left, right, 'add');
  return {
    ...left,
    basisPoints: left.basisPoints + right.basisPoints
  };
};
const subtractAmounts = (left, right) => {
  assertSameAmounts(left, right, 'subtract');
  return {
    ...left,
    basisPoints: left.basisPoints - right.basisPoints
  };
};
const multiplyAmount = (left, multiplier) => {
  if (typeof multiplier === 'bigint') {
    return {
      ...left,
      basisPoints: left.basisPoints * multiplier
    };
  }
  const [units, decimals] = multiplier.toString().split('.');
  const multiplierBasisPoints = BigInt(units + (decimals ?? ''));
  const multiplierExponents = BigInt(10) ** BigInt(decimals?.length ?? 0);
  return {
    ...left,
    basisPoints: left.basisPoints * multiplierBasisPoints / multiplierExponents
  };
};
const divideAmount = (left, divisor) => {
  if (typeof divisor === 'bigint') {
    return {
      ...left,
      basisPoints: left.basisPoints / divisor
    };
  }
  const [units, decimals] = divisor.toString().split('.');
  const divisorBasisPoints = BigInt(units + (decimals ?? ''));
  const divisorExponents = BigInt(10) ** BigInt(decimals?.length ?? 0);
  return {
    ...left,
    basisPoints: left.basisPoints * divisorExponents / divisorBasisPoints
  };
};
const absoluteAmount = value => {
  const x = value.basisPoints;
  return {
    ...value,
    basisPoints: x < 0 ? -x : x
  };
};
const compareAmounts = (left, right) => {
  assertSameAmounts(left, right, 'compare');
  if (left.basisPoints > right.basisPoints) return 1;
  if (left.basisPoints < right.basisPoints) return -1;
  return 0;
};
const isEqualToAmount = (left, right, tolerance) => {
  tolerance = tolerance ?? createAmount(0, left.identifier, left.decimals);
  assertSameAmounts(left, right, 'isEqualToAmount');
  assertSameAmounts(left, tolerance, 'isEqualToAmount');
  const delta = absoluteAmount(subtractAmounts(left, right));
  return isLessThanOrEqualToAmount(delta, tolerance);
};
const isLessThanAmount = (left, right) => compareAmounts(left, right) < 0;
const isLessThanOrEqualToAmount = (left, right) => compareAmounts(left, right) <= 0;
const isGreaterThanAmount = (left, right) => compareAmounts(left, right) > 0;
const isGreaterThanOrEqualToAmount = (left, right) => compareAmounts(left, right) >= 0;
const isZeroAmount = value => value.basisPoints === BigInt(0);
const isPositiveAmount = value => value.basisPoints >= BigInt(0);
const isNegativeAmount = value => value.basisPoints < BigInt(0);
const amountToString = (value, maxDecimals) => {
  let text = value.basisPoints.toString();
  if (value.decimals === 0) {
    return text;
  }
  const sign = text.startsWith('-') ? '-' : '';
  text = text.replace('-', '');
  text = text.padStart(value.decimals + 1, '0');
  const units = text.slice(0, -value.decimals);
  let decimals = text.slice(-value.decimals);
  if (maxDecimals !== undefined) {
    decimals = decimals.slice(0, maxDecimals);
  }
  return `${sign + units}.${decimals}`;
};
const amountToNumber = value => parseFloat(amountToString(value));
const displayAmount = (value, maxDecimals) => {
  const amountAsString = amountToString(value, maxDecimals);
  switch (value.identifier) {
    case '%':
      return `${amountAsString}%`;
    default:
      return `${value.identifier} ${amountAsString}`;
  }
};
const mapAmountSerializer = (serializer, identifier, decimals) => Serializer.mapSerializer(serializer, value => value.basisPoints > Number.MAX_SAFE_INTEGER ? value.basisPoints : Number(value.basisPoints), value => createAmount(value, identifier, decimals));

exports.absoluteAmount = absoluteAmount;
exports.addAmounts = addAmounts;
exports.amountToNumber = amountToNumber;
exports.amountToString = amountToString;
exports.assertAmount = assertAmount;
exports.assertSameAmounts = assertSameAmounts;
exports.assertSolAmount = assertSolAmount;
exports.compareAmounts = compareAmounts;
exports.createAmount = createAmount;
exports.createAmountFromDecimals = createAmountFromDecimals;
exports.displayAmount = displayAmount;
exports.divideAmount = divideAmount;
exports.isAmount = isAmount;
exports.isEqualToAmount = isEqualToAmount;
exports.isGreaterThanAmount = isGreaterThanAmount;
exports.isGreaterThanOrEqualToAmount = isGreaterThanOrEqualToAmount;
exports.isLessThanAmount = isLessThanAmount;
exports.isLessThanOrEqualToAmount = isLessThanOrEqualToAmount;
exports.isNegativeAmount = isNegativeAmount;
exports.isPositiveAmount = isPositiveAmount;
exports.isSolAmount = isSolAmount;
exports.isZeroAmount = isZeroAmount;
exports.lamports = lamports;
exports.mapAmountSerializer = mapAmountSerializer;
exports.multiplyAmount = multiplyAmount;
exports.sameAmounts = sameAmounts;
exports.sol = sol;
exports.subtractAmounts = subtractAmounts;
exports.tokenAmount = tokenAmount;
exports.usd = usd;
//# sourceMappingURL=Amount.cjs.map
